@@ -20,8 +20,10 @@ import os
 from chirp import util, chirp_common, bitwise, memmap, errors, directory
 from chirp.settings import RadioSetting, RadioSettingGroup, \
                 RadioSettingValueBoolean, RadioSettingValueList, \
-                RadioSettingValueInteger, RadioSettingValueString
+                RadioSettingValueInteger, RadioSettingValueString, \
+                RadioSettingValueFloat
 from chirp.wouxun_common import wipe_memory, do_download, do_upload
+from textwrap import dedent
 
 FREQ_ENCODE_TABLE = [ 0x7, 0xa, 0x0, 0x9, 0xb, 0x2, 0xe, 0x1, 0x3, 0xf ]
  
@@ -54,7 +56,7 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
     MODEL = "KG-UVD1P"
     _model = "KG669V"
     
-    _querymodel = "HiWOUXUN\x02"
+    _querymodel = ("HiWOUXUN\x02", "PROGUV6X\x02")
     
     CHARSET = list("0123456789") + [chr(x + ord("A")) for x in range(0, 26)] + \
         list("?+-")
@@ -83,6 +85,12 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
           u8 unknown[2];
         } memory[199];
         
+        #seekto 0x0842;
+        u16 fm_presets_0[9];
+
+        #seekto 0x0882;
+        u16 fm_presets_1[9];
+
         #seekto 0x0970;
         struct {
             u16 vhf_rx_start;
@@ -95,10 +103,86 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
             u16 uhf_tx_stop;
         } freq_ranges;
 
-        #seekto 0x0E5C;
+        #seekto 0x0E00;
         struct {
-          u8 unknown_flag1:7,
+          char welcome1[6];
+          char welcome2[6];
+          char single_band[6];
+        } strings;
+
+        #seekto 0x0E20;
+        struct {
+          u8 unknown_flag_01:6,
+             vfo_b_ch_disp:2;
+          u8 unknown_flag_02:5,
+             vfo_a_fr_step:3;
+          u8 unknown_flag_03:4,
+             vfo_a_squelch:4;
+          u8 unknown_flag_04:7,
+             power_save:1;
+          u8 unknown_flag_05:8;
+          u8 unknown_flag_06:6,
+             roger_beep:2;
+          u8 unknown_flag_07:2,
+             transmit_time_out:6;
+          u8 unknown_flag_08:4,
+             vox:4;
+          u8 unknown_1[4];
+          u8 unknown_flag_09:6,
+             voice:2;
+          u8 unknown_flag_10:7,
+             beep:1;
+          u8 unknown_flag_11:7,
+             ani_id_enable:1;
+          u8 unknown_2[2];
+          u8 unknown_flag_12:5,
+             vfo_b_fr_step:3;
+          u8 unknown_3[1];
+          u8 unknown_flag_13:3,
+             ani_id_tx_delay:5;
+          u8 unknown_4[1];
+          u8 unknown_flag_14:6,
+             ani_id_sidetone:2;
+          u8 unknown_flag_15:4,
+             tx_time_out_alert:4;
+          u8 unknown_flag_16:6,
+             vfo_a_ch_disp:2;
+          u8 unknown_flag_15:6,
+             scan_mode:2;
+          u8 unknown_flag_16:7,
+             kbd_lock:1;
+          u8 unknown_flag_17:6,
+             ponmsg:2;
+          u8 unknown_flag_18:5,
+             pf1_function:3;
+          u8 unknown_5[1];
+          u8 unknown_flag_19:7,
+             auto_backlight:1;
+          u8 unknown_flag_20:7,
+             sos_ch:1;
+          u8 unknown_6;
+          u8 sd_available;
+          u8 unknown_flag_21:7,
+             auto_lock_kbd:1;
+          u8 unknown_flag_22:4,
+             vfo_b_squelch:4;
+          u8 unknown_7[1];
+          u8 unknown_flag_23:7,
+             stopwatch:1;
+          u8 vfo_a_cur_chan;
+          u8 unknown_flag_24:7,
+             dual_band_receive:1;
+          u8 current_vfo:1,
+             unknown_flag_24:7;
+          u8 unknown_8[2];
+          u8 mode_password[6];
+          u8 reset_password[6];
+          u8 ani_id_content[6];
+          u8 unknown_flag_25:7,
              menu_available:1;
+          u8 unknown_9[1];
+          u8 priority_chan;
+          u8 vfo_b_cur_chan;
         } settings;
 
         #seekto 0x1008;
@@ -110,18 +194,46 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
     """
 
     @classmethod
-    def get_experimental_warning(cls):
-        return ('This version of the Wouxun driver allows you to modify the '
+    def get_prompts(cls):
+        rp = chirp_common.RadioPrompts()
+        rp.experimental = ('This version of the Wouxun driver allows you to modify the '
                 'frequency range settings of your radio. This has been tested '
                 'and reports from other users indicate that it is a safe '
                 'thing to do. However, modifications to this value may have '
                 'unintended consequences, including damage to your device. '
                 'You have been warned. Proceed at your own risk!')
+        rp.pre_download = _(dedent("""\
+            1. Turn radio off.
+            2. Connect cable to mic/spkr connector.
+            3. Make sure connector is firmly connected.
+            4. Turn radio on.
+            5. Ensure that the radio is tuned to channel with no activity.
+            6. Click OK to download image from device."""))
+        rp.pre_upload = _(dedent("""\
+            1. Turn radio off.
+            2. Connect cable to mic/spkr connector.
+            3. Make sure connector is firmly connected.
+            4. Turn radio on.
+            5. Ensure that the radio is tuned to channel with no activity.
+            6. Click OK to upload image to device."""))
+        return rp
+                
+    @classmethod
+    def _get_querymodel(cls):
+        if isinstance(cls._querymodel, str):
+            while True:
+                yield cls._querymodel
+        else:
+            i = 0
+            while True:
+                yield cls._querymodel[i % len(cls._querymodel)]
+                i += 1
 
     def _identify(self):
         """Do the original wouxun identification dance"""
-        for _i in range(0, 5):
-            self.pipe.write(self._querymodel)
+        query = self._get_querymodel()
+        for _i in range(0, 10):
+            self.pipe.write(query.next())
             resp = self.pipe.read(9)
             if len(resp) != 9:
                 print "Got:\n%s" % util.hexprint(resp)
@@ -218,50 +330,51 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
 
     def get_settings(self):
         freqranges = RadioSettingGroup("freqranges", "Freq ranges")
-        top = RadioSettingGroup("top", "All Settings", freqranges)
+        fm_preset = RadioSettingGroup("fm_preset", "FM Presets")
+        top = RadioSettingGroup("top", "All Settings", freqranges, fm_preset)
 
         rs = RadioSetting("menu_available", "Menu Available",
                           RadioSettingValueBoolean(
                             self._memobj.settings.menu_available))
         top.append(rs)
 
-        rs = RadioSetting("vhf_rx_start", "VHF RX Lower Limit (MHz)",
-                          RadioSettingValueInteger(136, 174, 
+        rs = RadioSetting("vhf_rx_start", "1st band RX Lower Limit (MHz)",
+                          RadioSettingValueInteger(50, 174, 
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_rx_start)))
         freqranges.append(rs)
-        rs = RadioSetting("vhf_rx_stop", "VHF RX Upper Limit (MHz)",
-                          RadioSettingValueInteger(136, 174, 
+        rs = RadioSetting("vhf_rx_stop", "1st band RX Upper Limit (MHz)",
+                          RadioSettingValueInteger(50, 174, 
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_rx_stop)))
         freqranges.append(rs)
-        rs = RadioSetting("uhf_rx_start", "UHF RX Lower Limit (MHz)",
-                          RadioSettingValueInteger(216, 520, 
+        rs = RadioSetting("uhf_rx_start", "2nd band RX Lower Limit (MHz)",
+                          RadioSettingValueInteger(136, 520, 
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_rx_start)))
         freqranges.append(rs)
-        rs = RadioSetting("uhf_rx_stop", "UHF RX Upper Limit (MHz)",
-                          RadioSettingValueInteger(216, 520, 
+        rs = RadioSetting("uhf_rx_stop", "2nd band RX Upper Limit (MHz)",
+                          RadioSettingValueInteger(136, 520, 
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_rx_stop)))
         freqranges.append(rs)
-        rs = RadioSetting("vhf_tx_start", "VHF TX Lower Limit (MHz)",
-                          RadioSettingValueInteger(136, 174, 
+        rs = RadioSetting("vhf_tx_start", "1st band TX Lower Limit (MHz)",
+                          RadioSettingValueInteger(50, 174, 
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_tx_start)))
         freqranges.append(rs)
-        rs = RadioSetting("vhf_tx_stop", "VHF TX Upper Limit (MHz)",
-                          RadioSettingValueInteger(136, 174, 
+        rs = RadioSetting("vhf_tx_stop", "1st TX Upper Limit (MHz)",
+                          RadioSettingValueInteger(50, 174, 
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_tx_stop)))
         freqranges.append(rs)
-        rs = RadioSetting("uhf_tx_start", "UHF TX Lower Limit (MHz)",
-                          RadioSettingValueInteger(216, 520, 
+        rs = RadioSetting("uhf_tx_start", "2st band TX Lower Limit (MHz)",
+                          RadioSettingValueInteger(136, 520, 
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_tx_start)))
         freqranges.append(rs)
-        rs = RadioSetting("uhf_tx_stop", "UHF TX Upper Limit (MHz)",
-                          RadioSettingValueInteger(216, 520, 
+        rs = RadioSetting("uhf_tx_stop", "2st band TX Upper Limit (MHz)",
+                          RadioSettingValueInteger(136, 520, 
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_tx_stop)))
         freqranges.append(rs)
@@ -269,19 +382,236 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
         # tell the decoded ranges to UI
         self.valid_freq = [
             ( decode_freq(self._memobj.freq_ranges.vhf_rx_start) * 1000000, 
-             (decode_freq(self._memobj.freq_ranges.vhf_rx_stop)+1) * 1000000), 
+             (decode_freq(self._memobj.freq_ranges.vhf_rx_stop)+1) * 1000000),
             ( decode_freq(self._memobj.freq_ranges.uhf_rx_start)  * 1000000,
              (decode_freq(self._memobj.freq_ranges.uhf_rx_stop)+1) * 1000000)]
 
+        def _filter(name):
+            filtered = ""
+            for char in str(name):
+                if char in chirp_common.CHARSET_ASCII:
+                    filtered += char
+                else:
+                    filtered += " "
+            return filtered
+
+        # add some radio specific settings
+        options = ["Off", "Welcome", "V bat"]
+        rs = RadioSetting("ponmsg", "Poweron message",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.ponmsg]))
+        top.append(rs)
+        rs = RadioSetting("strings.welcome1", "Power-On Message 1",
+                          RadioSettingValueString(0, 6,
+                              _filter(self._memobj.strings.welcome1)))
+        top.append(rs)
+        rs = RadioSetting("strings.welcome2", "Power-On Message 2",
+                          RadioSettingValueString(0, 6,
+                              _filter(self._memobj.strings.welcome2)))
+        top.append(rs)
+        rs = RadioSetting("strings.single_band", "Single Band Message",
+                          RadioSettingValueString(0, 6,
+                              _filter(self._memobj.strings.single_band)))
+        top.append(rs)
+        options = ["Channel", "ch/freq","Name", "VFO"]
+        rs = RadioSetting("vfo_a_ch_disp", "VFO A Channel disp mode",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.vfo_a_ch_disp]))
+        top.append(rs)
+        rs = RadioSetting("vfo_b_ch_disp", "VFO B Channel disp mode",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.vfo_b_ch_disp]))
+        top.append(rs)
+        options = ["5.0", "6.25", "10.0", "12.5", "25.0", "50.0", "100.0"]
+        rs = RadioSetting("vfo_a_fr_step", "VFO A Frequency Step",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.vfo_a_fr_step]))
+        top.append(rs)
+        rs = RadioSetting("vfo_b_fr_step", "VFO B Frequency Step",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.vfo_b_fr_step]))
+        top.append(rs)
+        rs = RadioSetting("vfo_a_squelch", "VFO A Squelch",
+                          RadioSettingValueInteger(0, 9,
+                              self._memobj.settings.vfo_a_squelch))
+        top.append(rs)
+        rs = RadioSetting("vfo_b_squelch", "VFO B Squelch",
+                          RadioSettingValueInteger(0, 9,
+                              self._memobj.settings.vfo_b_squelch))
+        top.append(rs)
+        rs = RadioSetting("vfo_a_cur_chan", "VFO A current channel",
+                          RadioSettingValueInteger(1, 128,
+                              self._memobj.settings.vfo_a_cur_chan))
+        top.append(rs)
+        rs = RadioSetting("vfo_b_cur_chan", "VFO B current channel",
+                          RadioSettingValueInteger(1, 128,
+                              self._memobj.settings.vfo_b_cur_chan))
+        top.append(rs)
+        rs = RadioSetting("priority_chan", "Priority channel",
+                          RadioSettingValueInteger(0, 199,
+                              self._memobj.settings.priority_chan))
+        top.append(rs)
+        rs = RadioSetting("power_save", "Power save",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.power_save))
+        top.append(rs)
+        options = ["Off", "Scan", "Lamp", "SOS", "Radio"]
+        rs = RadioSetting("pf1_function", "PF1 Function select",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.pf1_function]))
+        top.append(rs)
+        options = ["Off", "Begin", "End", "Both"]
+        rs = RadioSetting("roger_beep", "Roger beep select",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.roger_beep]))
+        top.append(rs)
+        options = ["%s" % x for x in range(15, 615, 15)]
+        rs = RadioSetting("transmit_time_out", "TX Time-out Timer",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.transmit_time_out]))
+        top.append(rs)
+        rs = RadioSetting("tx_time_out_alert", "TX Time-out Alert",
+                          RadioSettingValueInteger(0, 10,
+                              self._memobj.settings.tx_time_out_alert))
+        top.append(rs)
+        rs = RadioSetting("vox", "Vox",
+                          RadioSettingValueInteger(0, 10,
+                              self._memobj.settings.vox))
+        top.append(rs)
+        options = ["Off", "Chinese", "English"]
+        rs = RadioSetting("voice", "Voice",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.voice]))
+        top.append(rs)
+        rs = RadioSetting("beep", "Beep",
+                          RadioSettingValueBoolean(self._memobj.settings.beep))
+        top.append(rs)
+        rs = RadioSetting("ani_id_enable", "ANI id enable",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.ani_id_enable))
+        top.append(rs)
+        rs = RadioSetting("ani_id_tx_delay", "ANI id tx delay",
+                          RadioSettingValueInteger(0, 30,
+                              self._memobj.settings.ani_id_tx_delay))
+        top.append(rs)
+        options = ["Off", "Key", "ANI", "Key+ANI"]
+        rs = RadioSetting("ani_id_sidetone", "ANI id sidetone",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.ani_id_sidetone]))
+        top.append(rs)
+        options = ["Time", "Carrier", "Search"]
+        rs = RadioSetting("scan_mode", "Scan mode",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.scan_mode]))
+        top.append(rs)
+        rs = RadioSetting("kbd_lock", "Keyboard lock",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.kbd_lock))
+        top.append(rs)
+        rs = RadioSetting("auto_lock_kbd", "Auto lock keyboard",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.auto_lock_kbd))
+        top.append(rs)
+        rs = RadioSetting("auto_backlight", "Auto backlight",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.auto_backlight))
+        top.append(rs)
+        options = ["CH A", "CH B"]
+        rs = RadioSetting("sos_ch", "SOS CH",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.sos_ch]))
+        top.append(rs)
+        rs = RadioSetting("stopwatch", "Stopwatch",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.stopwatch))
+        top.append(rs)
+        rs = RadioSetting("dual_band_receive", "Dual band receive",
+                          RadioSettingValueBoolean(
+                              self._memobj.settings.dual_band_receive))
+        top.append(rs)
+        options = ["VFO A", "VFO B"]
+        rs = RadioSetting("current_vfo", "Current VFO",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.current_vfo]))
+        top.append(rs)
+
+        options = ["Dual", "Single"]
+        rs = RadioSetting("sd_available", "Single/Dual Band",
+                          RadioSettingValueList(options,
+                              options[self._memobj.settings.sd_available]))
+        top.append(rs)
+
+        _pwd = self._memobj.settings.mode_password
+        rs = RadioSetting("mode_password", "Mode password (000000 disabled)",
+                  RadioSettingValueInteger(0, 9, _pwd[0]),
+                  RadioSettingValueInteger(0, 9, _pwd[1]),
+                  RadioSettingValueInteger(0, 9, _pwd[2]),
+                  RadioSettingValueInteger(0, 9, _pwd[3]),
+                  RadioSettingValueInteger(0, 9, _pwd[4]),
+                  RadioSettingValueInteger(0, 9, _pwd[5]))
+        top.append(rs)
+        _pwd = self._memobj.settings.reset_password
+        rs = RadioSetting("reset_password", "Reset password (000000 disabled)",
+                  RadioSettingValueInteger(0, 9, _pwd[0]),
+                  RadioSettingValueInteger(0, 9, _pwd[1]),
+                  RadioSettingValueInteger(0, 9, _pwd[2]),
+                  RadioSettingValueInteger(0, 9, _pwd[3]),
+                  RadioSettingValueInteger(0, 9, _pwd[4]),
+                  RadioSettingValueInteger(0, 9, _pwd[5]))
+        top.append(rs)
+
+        dtmfchars = "0123456789 *#ABCD"
+        _codeobj = self._memobj.settings.ani_id_content
+        _code = "".join([dtmfchars[x] for x in _codeobj if int(x) < 0x1F])
+        val = RadioSettingValueString(0, 6, _code, False)
+        val.set_charset(dtmfchars)
+        rs = RadioSetting("settings.ani_id_content", "PTT-ID Code", val)
+        def apply_ani_id(setting, obj):
+            value = []
+            for j in range(0, 6):
+                try:
+                    value.append(dtmfchars.index(str(setting.value)[j]))
+                except IndexError:
+                    value.append(0xFF)
+            obj.ani_id_content = value
+        rs.set_apply_callback(apply_ani_id, self._memobj.settings)
+        top.append(rs)
+
+        for i in range(0, 9):
+            if self._memobj.fm_presets_0[i] != 0xFFFF:
+                used = True
+                preset = self._memobj.fm_presets_0[i]/10.0+76
+            else:
+                used = False
+                preset = 76
+            rs = RadioSetting("fm_presets_0_%1i" % i, "Team 1 Location %i" % (i+1),
+                          RadioSettingValueBoolean(used),
+                          RadioSettingValueFloat(76, 108, preset, 0.1, 1))
+            fm_preset.append(rs)
+        for i in range(0, 9):
+            if self._memobj.fm_presets_1[i] != 0xFFFF:
+                used = True
+                preset = self._memobj.fm_presets_1[i]/10.0+76
+            else:
+                used = False
+                preset = 76
+            rs = RadioSetting("fm_presets_1_%1i" % i, "Team 2 Location %i" % (i+1),
+                          RadioSettingValueBoolean(used),
+                          RadioSettingValueFloat(76, 108, preset, 0.1, 1))
+            fm_preset.append(rs)
+
         return top
 
     def set_settings(self, settings):
         for element in settings:
             if not isinstance(element, RadioSetting):
-                if element.get_name() != "freqranges" :
-                    self.set_settings(element)
-                else:
+                if element.get_name() == "freqranges" :
                     self._set_freq_settings(element)
+                elif element.get_name() == "fm_preset" :
+                    self._set_fm_preset(element)
+                else:
+                    self.set_settings(element)
+                    continue
             else:
                 try:
                     if "." in element.get_name():
@@ -293,12 +623,37 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
                     else:
                         obj = self._memobj.settings
                         setting = element.get_name()
-                    print "Setting %s = %s" % (setting, element.value)
-                    setattr(obj, setting, element.value)
+
+                    if element.has_apply_callback():
+                        print "Using apply callback"
+                        element.run_apply_callback()
+                    else:
+                        print "Setting %s = %s" % (setting, element.value)
+                        setattr(obj, setting, element.value)
                 except Exception, e:
                     print element.get_name()
                     raise
 
+    def _set_fm_preset(self, settings):
+        obj = self._memobj
+        for element in settings:
+            try:
+                (bank, index) = (int(a) for a in element.get_name().split("_")[-2:])
+                val = element.value
+                if val[0].get_value():
+                    value = int(val[1].get_value()*10-760)
+                else:
+                    value = 0xffff
+                print "Setting fm_presets_%1i[%1i] = %s" % (bank, index, value)
+                if bank == 0:
+                    setting = self._memobj.fm_presets_0
+                else:
+                    setting = self._memobj.fm_presets_1
+                setting[index] = value
+            except Exception, e:
+                print element.get_name()
+                raise
+
     def _set_freq_settings(self, settings):
         for element in settings:
             try:
@@ -317,7 +672,8 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
             code = int("%03o" % (val & 0x07FF))
             pol = (val & 0x8000) and "R" or "N"
             return code, pol
-            
+
+        tpol = False
         if _mem.tx_tone != 0xFFFF and _mem.tx_tone > 0x2800:
             tcode, tpol = _get_dcs(_mem.tx_tone)
             mem.dtcs = tcode
@@ -328,6 +684,7 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
         else:
             txmode = ""
 
+        rpol = False
         if _mem.rx_tone != 0xFFFF and _mem.rx_tone > 0x2800:
             rcode, rpol = _get_dcs(_mem.rx_tone)
             mem.rx_dtcs = rcode
@@ -348,8 +705,8 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
             mem.tmode = "Cross"
             mem.cross_mode = "%s->%s" % (txmode, rxmode)
 
-        if mem.tmode == "DTCS":
-            mem.dtcs_polarity = "%s%s" % (tpol, rpol)
+        # always set it even if no dtcs is used
+        mem.dtcs_polarity = "%s%s" % (tpol or "N", rpol or "N")
 
         if os.getenv("CHIRP_DEBUG"):
             print "Got TX %s (%i) RX %s (%i)" % (txmode, _mem.tx_tone,
@@ -415,7 +772,7 @@ class KGUVD1PRadio(chirp_common.CloneModeRadio,
         def _set_dcs(code, pol):
             val = int("%i" % code, 8) + 0x2800
             if pol == "R":
-                val += 0xA000
+                val += 0x8000
             return val
 
         if mem.tmode == "Cross":
@@ -513,7 +870,7 @@ class KGUV6DRadio(KGUVD1PRadio):
     """Wouxun KG-UV6 (D and X variants)"""
     MODEL = "KG-UV6"
     
-    _querymodel = "HiWXUVD1\x02"
+    _querymodel = ("HiWXUVD1\x02", "HiKGUVD1\x02")
     
     _MEM_FORMAT = """
         #seekto 0x0010;
@@ -591,7 +948,8 @@ class KGUV6DRadio(KGUVD1PRadio):
              auto_backlight:1;
           u8 unknown_flag_20:7,
              sos_ch:1;
-          u8 unknown_6[2];
+          u8 unknown_6;
+          u8 sd_available;
           u8 unknown_flag_21:7,
              auto_lock_kbd:1;
           u8 unknown_flag_22:4,
@@ -632,7 +990,7 @@ class KGUV6DRadio(KGUVD1PRadio):
           u8 pad[2];
         } vfo_settings[2];
 	
-        #seekto 0x0f80;
+        #seekto 0x0f82;
         u16 fm_presets_0[9];
 
         #seekto 0x0ff0;
@@ -661,7 +1019,7 @@ class KGUV6DRadio(KGUVD1PRadio):
             u8 pad[9];
         } vfo_offset[2];
 
-        #seekto 0x1f80;
+        #seekto 0x1f82;
         u16 fm_presets_1[9];
     """
 
@@ -673,7 +1031,8 @@ class KGUV6DRadio(KGUVD1PRadio):
 
     def get_settings(self):
         freqranges = RadioSettingGroup("freqranges", "Freq ranges")
-        top = RadioSettingGroup("top", "All Settings", freqranges)
+        fm_preset = RadioSettingGroup("fm_preset", "FM Presets")
+        top = RadioSettingGroup("top", "All Settings", freqranges, fm_preset)
 
         rs = RadioSetting("menu_available", "Menu Available",
                           RadioSettingValueBoolean(
@@ -761,8 +1120,15 @@ class KGUV6DRadio(KGUVD1PRadio):
                           RadioSettingValueList(options,
                                         options[self._memobj.settings.vfo_b_ch_disp]))
         top.append(rs)
-	# TODO - vfo_a_fr_step
-	# TODO -vfo_b_fr_step:3;
+        options = ["2.5", "5.0", "6.25", "10.0", "12.5", "25.0", "50.0", "100.0"]
+        rs = RadioSetting("vfo_a_fr_step", "VFO A Frequency Step",
+                          RadioSettingValueList(options,
+                                        options[self._memobj.settings.vfo_a_fr_step]))
+        top.append(rs)
+        rs = RadioSetting("vfo_b_fr_step", "VFO B Frequency Step",
+                          RadioSettingValueList(options,
+                                        options[self._memobj.settings.vfo_b_fr_step]))
+        top.append(rs)
         rs = RadioSetting("vfo_a_squelch", "VFO A Squelch",
                           RadioSettingValueInteger(0, 9, self._memobj.settings.vfo_a_squelch))
         top.append(rs)
@@ -773,7 +1139,7 @@ class KGUV6DRadio(KGUVD1PRadio):
                           RadioSettingValueInteger(1, 199, self._memobj.settings.vfo_a_cur_chan))
         top.append(rs)
         rs = RadioSetting("vfo_b_cur_chan", "VFO B current channel",
-                          RadioSettingValueInteger(0, 199, self._memobj.settings.vfo_b_cur_chan))
+                          RadioSettingValueInteger(1, 199, self._memobj.settings.vfo_b_cur_chan))
         top.append(rs)
         rs = RadioSetting("priority_chan", "Priority channel",
                           RadioSettingValueInteger(0, 199, self._memobj.settings.priority_chan))
@@ -796,7 +1162,14 @@ class KGUV6DRadio(KGUVD1PRadio):
                           RadioSettingValueList(options,
                                         options[self._memobj.settings.roger_beep]))
         top.append(rs)
-        # TODO - transmit_time_out:6;
+        options = ["%s" % x for x in range(15, 615, 15)]
+        rs = RadioSetting("transmit_time_out", "TX Time-out Timer",
+                          RadioSettingValueList(options,
+                                        options[self._memobj.settings.transmit_time_out]))
+        top.append(rs)
+        rs = RadioSetting("tx_time_out_alert", "TX Time-out Alert",
+                          RadioSettingValueInteger(0, 10, self._memobj.settings.tx_time_out_alert))
+        top.append(rs)
         rs = RadioSetting("vox", "Vox",
                           RadioSettingValueInteger(0, 10, self._memobj.settings.vox))
         top.append(rs)
@@ -819,7 +1192,6 @@ class KGUV6DRadio(KGUVD1PRadio):
                           RadioSettingValueList(options,
                                         options[self._memobj.settings.ani_id_sidetone]))
         top.append(rs)
-        # TODO tx_time_out_alert:4;
         options = ["Time", "Carrier", "Search"]
         rs = RadioSetting("scan_mode", "Scan mode",
                           RadioSettingValueList(options,
@@ -850,6 +1222,13 @@ class KGUV6DRadio(KGUVD1PRadio):
                           RadioSettingValueList(options,
                                         options[self._memobj.settings.current_vfo]))
         top.append(rs)
+
+        options = ["Dual", "Single"]
+        rs = RadioSetting("sd_available", "Single/Dual Band",
+                          RadioSettingValueList(options,
+                                        options[self._memobj.settings.sd_available]))
+        top.append(rs)
+
         _pwd = self._memobj.settings.mode_password
         rs = RadioSetting("mode_password", "Mode password (000000 disabled)",
                   RadioSettingValueInteger(0, 9, _pwd[0]),
@@ -868,22 +1247,100 @@ class KGUV6DRadio(KGUVD1PRadio):
                   RadioSettingValueInteger(0, 9, _pwd[4]),
                   RadioSettingValueInteger(0, 9, _pwd[5]))
         top.append(rs)
-        try:
-            _ani = self._memobj.settings.ani_id_content
-            rs = RadioSetting("ani_id_content", "ANI Code",
-                              RadioSettingValueInteger(0, 9, _ani[0]),
-                              RadioSettingValueInteger(0, 9, _ani[1]),
-                              RadioSettingValueInteger(0, 9, _ani[2]),
-                              RadioSettingValueInteger(0, 9, _ani[3]),
-                              RadioSettingValueInteger(0, 9, _ani[4]),
-                              RadioSettingValueInteger(0, 9, _ani[5]))
-            top.append(rs)
-        except Exception:
-            print ("Your ANI code is not five digits, which is not currently"
-                   " supported in CHIRP.")            
+
+        dtmfchars = "0123456789 *#ABCD"
+        _codeobj = self._memobj.settings.ani_id_content
+        _code = "".join([dtmfchars[x] for x in _codeobj if int(x) < 0x1F])
+        val = RadioSettingValueString(0, 6, _code, False)
+        val.set_charset(dtmfchars)
+        rs = RadioSetting("settings.ani_id_content", "ANI Code", val)
+        def apply_ani_id(setting, obj):
+            value = []
+            for j in range(0, 6):
+                try:
+                    value.append(dtmfchars.index(str(setting.value)[j]))
+                except IndexError:
+                    value.append(0xFF)
+            obj.ani_id_content = value
+        rs.set_apply_callback(apply_ani_id, self._memobj.settings)
+        top.append(rs)
+
+        for i in range(0, 9):
+            if self._memobj.fm_presets_0[i] != 0xFFFF:
+                used = True
+                preset = self._memobj.fm_presets_0[i]/10.0+76
+            else:
+                used = False
+                preset = 76
+            rs = RadioSetting("fm_presets_0_%1i" % i, "Team 1 Location %i" % (i+1),
+                          RadioSettingValueBoolean(used),
+                          RadioSettingValueFloat(76, 108, preset, 0.1, 1))
+            fm_preset.append(rs)
+        for i in range(0, 9):
+            if self._memobj.fm_presets_1[i] != 0xFFFF:
+                used = True
+                preset = self._memobj.fm_presets_1[i]/10.0+76
+            else:
+                used = False
+                preset = 76
+            rs = RadioSetting("fm_presets_1_%1i" % i, "Team 2 Location %i" % (i+1),
+                          RadioSettingValueBoolean(used),
+                          RadioSettingValueFloat(76, 108, preset, 0.1, 1))
+            fm_preset.append(rs)
 
         return top
 
+    def set_settings(self, settings):
+        for element in settings:
+            if not isinstance(element, RadioSetting):
+                if element.get_name() == "freqranges" :
+                    self._set_freq_settings(element)
+                elif element.get_name() == "fm_preset" :
+                    self._set_fm_preset(element)
+                else:
+                    self.set_settings(element)
+            else:
+                try:
+                    if "." in element.get_name():
+                        bits = element.get_name().split(".")
+                        obj = self._memobj
+                        for bit in bits[:-1]:
+                            obj = getattr(obj, bit)
+                        setting = bits[-1]
+                    else:
+                        obj = self._memobj.settings
+                        setting = element.get_name()
+
+                    if element.has_apply_callback():
+                        print "Using apply callback"
+                        element.run_apply_callback()
+                    else:
+                        print "Setting %s = %s" % (setting, element.value)
+                        setattr(obj, setting, element.value)
+                except Exception, e:
+                    print element.get_name()
+                    raise
+
+    def _set_fm_preset(self, settings):
+        obj = self._memobj
+        for element in settings:
+            try:
+                (bank, index) = (int(a) for a in element.get_name().split("_")[-2:])
+                val = element.value
+                if val[0].get_value():
+                    value = int(val[1].get_value()*10-760)
+                else:
+                    value = 0xffff
+                print "Setting fm_presets_%1i[%1i] = %s" % (bank, index, value)
+                if bank == 0:
+                    setting = self._memobj.fm_presets_0
+                else:
+                    setting = self._memobj.fm_presets_1
+                setting[index] = value
+            except Exception, e:
+                print element.get_name()
+                raise
+
     @classmethod
     def match_model(cls, filedata, filename):
         if len(filedata) == 8192 and \
@@ -894,9 +1351,11 @@ class KGUV6DRadio(KGUVD1PRadio):
 @directory.register
 class KG816Radio(KGUVD1PRadio,
         chirp_common.ExperimentalRadio):
-    """Wouxun KG816"""
-    MODEL = "KG816"
+    """Wouxun KG-816"""
+    MODEL = "KG-816"
 
+    _querymodel = "HiWOUXUN\x02"
+    
     _MEM_FORMAT = """
         #seekto 0x0010;
         struct {
@@ -953,42 +1412,42 @@ class KG816Radio(KGUVD1PRadio,
         top = RadioSettingGroup("top", "All Settings", freqranges)
 
         rs = RadioSetting("vhf_rx_start", "vhf rx start",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_rx_start)))
         freqranges.append(rs)
         rs = RadioSetting("vhf_rx_stop", "vhf rx stop",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_rx_stop)))
         freqranges.append(rs)
         rs = RadioSetting("uhf_rx_start", "uhf rx start",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_rx_start)))
         freqranges.append(rs)
         rs = RadioSetting("uhf_rx_stop", "uhf rx stop",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_rx_stop)))
         freqranges.append(rs)
         rs = RadioSetting("vhf_tx_start", "vhf tx start",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_tx_start)))
         freqranges.append(rs)
         rs = RadioSetting("vhf_tx_stop", "vhf tx stop",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.vhf_tx_stop)))
         freqranges.append(rs)
         rs = RadioSetting("uhf_tx_start", "uhf tx start",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_tx_start)))
         freqranges.append(rs)
         rs = RadioSetting("uhf_tx_stop", "uhf tx stop",
-                          RadioSettingValueInteger(136, 520, 
+                          RadioSettingValueInteger(66, 520,
                                 decode_freq(
                                     self._memobj.freq_ranges.uhf_tx_stop)))
         freqranges.append(rs)
@@ -1010,3 +1469,12 @@ class KG816Radio(KGUVD1PRadio,
             return True
         return False
 
+
+@directory.register
+class KG818Radio(KG816Radio):
+    """Wouxun KG-818"""
+    MODEL = "KG-818"
+
+    @classmethod
+    def match_model(cls, filedata, filename):
+        return False
